@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2004-2009, 2011-2017  Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (C) 2004-2009, 2011-2015  Internet Systems Consortium, Inc. ("ISC")
  * Copyright (C) 2000-2003  Internet Software Consortium.
  *
  * Permission to use, copy, modify, and/or distribute this software for any
@@ -17,6 +17,7 @@
 
 /*
  * Principal Author: Brian Wellington
+ * $Id$
  */
 #ifdef OPENSSL
 #include <config.h>
@@ -105,8 +106,7 @@
 	(rsa)->flags &= ~(RSA_FLAG_CACHE_PUBLIC | RSA_FLAG_CACHE_PRIVATE); \
 	(rsa)->flags &= ~RSA_FLAG_BLINDING; \
 	} while (0)
-#elif OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
-#if defined(RSA_FLAG_NO_BLINDING)
+#elif defined(RSA_FLAG_NO_BLINDING)
 #define SET_FLAGS(rsa) \
 	do { \
 		(rsa)->flags &= ~RSA_FLAG_BLINDING; \
@@ -118,132 +118,9 @@
 		(rsa)->flags &= ~RSA_FLAG_BLINDING; \
 	} while (0)
 #endif
-#else
-#define SET_FLAGS(rsa) \
-	do { \
-		RSA_clear_flags(rsa, RSA_FLAG_BLINDING); \
-		RSA_set_flags(rsa, RSA_FLAG_NO_BLINDING); \
-	} while (0)
-#endif
+
 #define DST_RET(a) {ret = a; goto err;}
 
-#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
-/* From OpenSSL 1.1.0 */
-static int
-RSA_set0_key(RSA *r, BIGNUM *n, BIGNUM *e, BIGNUM *d) {
-
-	/*
-	 * If the fields n and e in r are NULL, the corresponding input
-	 * parameters MUST be non-NULL for n and e.  d may be
-	 * left NULL (in case only the public key is used).
-	 */
-	if ((r->n == NULL && n == NULL) || (r->e == NULL && e == NULL))
-		return 0;
-
-	if (n != NULL) {
-		BN_free(r->n);
-		r->n = n;
-	}
-	if (e != NULL) {
-		BN_free(r->e);
-		r->e = e;
-	}
-	if (d != NULL) {
-		BN_free(r->d);
-		r->d = d;
-	}
-
-	return 1;
-}
-
-static int
-RSA_set0_factors(RSA *r, BIGNUM *p, BIGNUM *q) {
-
-	/*
-	 * If the fields p and q in r are NULL, the corresponding input
-	 * parameters MUST be non-NULL.
-	 */
-	if ((r->p == NULL && p == NULL) || (r->q == NULL && q == NULL))
-		return 0;
-
-	if (p != NULL) {
-		BN_free(r->p);
-		r->p = p;
-	}
-	if (q != NULL) {
-		BN_free(r->q);
-		r->q = q;
-	}
-
-	return 1;
-}
-
-static int
-RSA_set0_crt_params(RSA *r, BIGNUM *dmp1, BIGNUM *dmq1, BIGNUM *iqmp) {
-	/*
-	 * If the fields dmp1, dmq1 and iqmp in r are NULL, the
-	 * corresponding input parameters MUST be non-NULL.
-	 */
-	if ((r->dmp1 == NULL && dmp1 == NULL) ||
-	    (r->dmq1 == NULL && dmq1 == NULL) ||
-	    (r->iqmp == NULL && iqmp == NULL))
-		return 0;
-
-	if (dmp1 != NULL) {
-		BN_free(r->dmp1);
-		r->dmp1 = dmp1;
-	}
-	if (dmq1 != NULL) {
-		BN_free(r->dmq1);
-		r->dmq1 = dmq1;
-	}
-	if (iqmp != NULL) {
-		BN_free(r->iqmp);
-		r->iqmp = iqmp;
-	}
-
-	return 1;
-}
-
-static void
-RSA_get0_key(const RSA *r,
-	     const BIGNUM **n, const BIGNUM **e, const BIGNUM **d)
-{
-	if (n != NULL)
-		*n = r->n;
-	if (e != NULL)
-		*e = r->e;
-	if (d != NULL)
-		*d = r->d;
-}
-
-static void
-RSA_get0_factors(const RSA *r, const BIGNUM **p, const BIGNUM **q) {
-	if (p != NULL)
-		*p = r->p;
-	if (q != NULL)
-	*q = r->q;
-}
-
-static void
-RSA_get0_crt_params(const RSA *r, const BIGNUM **dmp1, const BIGNUM **dmq1,
-		    const BIGNUM **iqmp)
-{
-	if (dmp1 != NULL)
-		*dmp1 = r->dmp1;
-	if (dmq1 != NULL)
-		*dmq1 = r->dmq1;
-	if (iqmp != NULL)
-		*iqmp = r->iqmp;
-}
-
-static int
-RSA_test_flags(const RSA *r, int flags) {
-	return (r->flags & flags);
-}
-
-#endif
-
 static isc_result_t opensslrsa_todns(const dst_key_t *key, isc_buffer_t *data);
 
 static isc_result_t
@@ -639,7 +516,6 @@ static isc_result_t
 opensslrsa_verify2(dst_context_t *dctx, int maxbits, const isc_region_t *sig) {
 	dst_key_t *key = dctx->key;
 	int status = 0;
-	const BIGNUM *e = NULL;
 #if USE_EVP
 	EVP_MD_CTX *evp_md_ctx = dctx->ctxdata.evp_md_ctx;
 	EVP_PKEY *pkey = key->keydata.pkey;
@@ -667,8 +543,7 @@ opensslrsa_verify2(dst_context_t *dctx, int maxbits, const isc_region_t *sig) {
 	rsa = EVP_PKEY_get1_RSA(pkey);
 	if (rsa == NULL)
 		return (dst__openssl_toresult(DST_R_OPENSSLFAILURE));
-	RSA_get0_key(rsa, NULL, &e, NULL);
-	bits = BN_num_bits(e);
+	bits = BN_num_bits(rsa->e);
 	RSA_free(rsa);
 	if (bits > maxbits && maxbits != 0)
 		return (DST_R_VERIFYFAILURE);
@@ -685,8 +560,7 @@ opensslrsa_verify2(dst_context_t *dctx, int maxbits, const isc_region_t *sig) {
 					       DST_R_VERIFYFAILURE));
 	}
 #else
-	RSA_get0_key(rsa, NULL, &e, NULL);
-	if (BN_num_bits(e) > maxbits && maxbits != 0)
+	if (BN_num_bits(rsa->e) > maxbits && maxbits != 0)
 		return (DST_R_VERIFYFAILURE);
 
 	switch (dctx->key->key_alg) {
@@ -811,11 +685,6 @@ static isc_boolean_t
 opensslrsa_compare(const dst_key_t *key1, const dst_key_t *key2) {
 	int status;
 	RSA *rsa1 = NULL, *rsa2 = NULL;
-	const BIGNUM *n1 = NULL, *n2 = NULL;
-	const BIGNUM *e1 = NULL, *e2 = NULL;
-	const BIGNUM *d1 = NULL, *d2 = NULL;
-	const BIGNUM *p1 = NULL, *p2 = NULL;
-	const BIGNUM *q1 = NULL, *q2 = NULL;
 #if USE_EVP
 	EVP_PKEY *pkey1, *pkey2;
 #endif
@@ -845,18 +714,17 @@ opensslrsa_compare(const dst_key_t *key1, const dst_key_t *key2) {
 	else if (rsa1 == NULL || rsa2 == NULL)
 		return (ISC_FALSE);
 
-	RSA_get0_key(rsa1, &n1, &e1, &d1);
-	RSA_get0_key(rsa2, &n2, &e2, &d2);
-	status = BN_cmp(n1, n2) || BN_cmp(e1, e2);
+	status = BN_cmp(rsa1->n, rsa2->n) ||
+		 BN_cmp(rsa1->e, rsa2->e);
 
 	if (status != 0)
 		return (ISC_FALSE);
 
 #if USE_EVP
-	if (RSA_test_flags(rsa1, RSA_FLAG_EXT_PKEY) != 0 ||
-	    RSA_test_flags(rsa2, RSA_FLAG_EXT_PKEY) != 0) {
-		if (RSA_test_flags(rsa1, RSA_FLAG_EXT_PKEY) == 0 ||
-		    RSA_test_flags(rsa2, RSA_FLAG_EXT_PKEY) == 0)
+	if ((rsa1->flags & RSA_FLAG_EXT_PKEY) != 0 ||
+	    (rsa2->flags & RSA_FLAG_EXT_PKEY) != 0) {
+		if ((rsa1->flags & RSA_FLAG_EXT_PKEY) == 0 ||
+		    (rsa2->flags & RSA_FLAG_EXT_PKEY) == 0)
 			return (ISC_FALSE);
 		/*
 		 * Can't compare private parameters, BTW does it make sense?
@@ -865,12 +733,12 @@ opensslrsa_compare(const dst_key_t *key1, const dst_key_t *key2) {
 	}
 #endif
 
-	if (d1 != NULL || d2 != NULL) {
-		if (d1 == NULL || d2 == NULL)
+	if (rsa1->d != NULL || rsa2->d != NULL) {
+		if (rsa1->d == NULL || rsa2->d == NULL)
 			return (ISC_FALSE);
-		RSA_get0_factors(rsa1, &p1, &q1);
-		RSA_get0_factors(rsa2, &p2, &q2);
-		status = BN_cmp(d1, d2) || BN_cmp(p1, p1) || BN_cmp(q1, q2);
+		status = BN_cmp(rsa1->d, rsa2->d) ||
+			 BN_cmp(rsa1->p, rsa2->p) ||
+			 BN_cmp(rsa1->q, rsa2->q);
 
 		if (status != 0)
 			return (ISC_FALSE);
@@ -905,7 +773,7 @@ opensslrsa_generate(dst_key_t *key, int exp, void (*callback)(int)) {
 	} u;
 	RSA *rsa = RSA_new();
 	BIGNUM *e = BN_new();
-#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
 	BN_GENCB _cb;
 #endif
 	BN_GENCB *cb = BN_GENCB_new();
@@ -956,7 +824,7 @@ opensslrsa_generate(dst_key_t *key, int exp, void (*callback)(int)) {
 	ret = dst__openssl_toresult2("RSA_generate_key_ex",
 				     DST_R_OPENSSLFAILURE);
 
- err:
+err:
 #if USE_EVP
 	if (pkey != NULL)
 		EVP_PKEY_free(pkey);
@@ -1013,7 +881,6 @@ opensslrsa_generate(dst_key_t *key, int exp, void (*callback)(int)) {
 
 static isc_boolean_t
 opensslrsa_isprivate(const dst_key_t *key) {
-	const BIGNUM *d = NULL;
 #if USE_EVP
 	RSA *rsa = EVP_PKEY_get1_RSA(key->keydata.pkey);
 	INSIST(rsa != NULL);
@@ -1022,10 +889,9 @@ opensslrsa_isprivate(const dst_key_t *key) {
 #else
 	RSA *rsa = key->keydata.rsa;
 #endif
-	if (rsa != NULL && RSA_test_flags(rsa, RSA_FLAG_EXT_PKEY) != 0)
+	if (rsa != NULL && (rsa->flags & RSA_FLAG_EXT_PKEY) != 0)
 		return (ISC_TRUE);
-	RSA_get0_key(rsa, NULL, NULL, &d);
-	return (ISC_TF(rsa != NULL && d != NULL));
+	return (ISC_TF(rsa != NULL && rsa->d != NULL));
 }
 
 static void
@@ -1041,6 +907,7 @@ opensslrsa_destroy(dst_key_t *key) {
 #endif
 }
 
+
 static isc_result_t
 opensslrsa_todns(const dst_key_t *key, isc_buffer_t *data) {
 	isc_region_t r;
@@ -1051,7 +918,6 @@ opensslrsa_todns(const dst_key_t *key, isc_buffer_t *data) {
 #if USE_EVP
 	EVP_PKEY *pkey;
 #endif
-	const BIGNUM *e = NULL, *n = NULL;
 
 #if USE_EVP
 	REQUIRE(key->keydata.pkey != NULL);
@@ -1070,9 +936,8 @@ opensslrsa_todns(const dst_key_t *key, isc_buffer_t *data) {
 
 	isc_buffer_availableregion(data, &r);
 
-	RSA_get0_key(rsa, &n, &e, NULL);
-	mod_bytes = BN_num_bytes(n);
-	e_bytes = BN_num_bytes(e);
+	e_bytes = BN_num_bytes(rsa->e);
+	mod_bytes = BN_num_bytes(rsa->n);
 
 	if (e_bytes < 256) {	/*%< key exponent is <= 2040 bits */
 		if (r.length < 1)
@@ -1090,10 +955,9 @@ opensslrsa_todns(const dst_key_t *key, isc_buffer_t *data) {
 	if (r.length < e_bytes + mod_bytes)
 		DST_RET(ISC_R_NOSPACE);
 
-	RSA_get0_key(rsa, &n, &e, NULL);
-	BN_bn2bin(e, r.base);
+	BN_bn2bin(rsa->e, r.base);
 	isc_region_consume(&r, e_bytes);
-	BN_bn2bin(n, r.base);
+	BN_bn2bin(rsa->n, r.base);
 
 	isc_buffer_add(data, e_bytes + mod_bytes);
 
@@ -1115,7 +979,6 @@ opensslrsa_fromdns(dst_key_t *key, isc_buffer_t *data) {
 #if USE_EVP
 	EVP_PKEY *pkey;
 #endif
-	BIGNUM *e = NULL, *n = NULL;
 
 	isc_buffer_remainingregion(data, &r);
 	if (r.length == 0)
@@ -1149,16 +1012,12 @@ opensslrsa_fromdns(dst_key_t *key, isc_buffer_t *data) {
 		RSA_free(rsa);
 		return (DST_R_INVALIDPUBLICKEY);
 	}
-	e = BN_bin2bn(r.base, e_bytes, NULL);
+	rsa->e = BN_bin2bn(r.base, e_bytes, NULL);
 	isc_region_consume(&r, e_bytes);
-	n = BN_bin2bn(r.base, r.length, NULL);
-	if (RSA_set0_key(rsa, n, e, NULL) == 0) {
-		if (n != NULL) BN_free(n);
-		if (e != NULL) BN_free(e);
-		RSA_free(rsa);
-		return (ISC_R_NOMEMORY);
-	}
-	key->key_size = BN_num_bits(n);
+
+	rsa->n = BN_bin2bn(r.base, r.length, NULL);
+
+	key->key_size = BN_num_bits(rsa->n);
 
 	isc_buffer_forward(data, length);
 
@@ -1189,9 +1048,6 @@ opensslrsa_tofile(const dst_key_t *key, const char *directory) {
 	dst_private_t priv;
 	unsigned char *bufs[8];
 	isc_result_t result;
-	const BIGNUM *n = NULL, *e = NULL, *d = NULL;
-	const BIGNUM *p = NULL, *q = NULL;
-	const BIGNUM *dmp1 = NULL, *dmq1 = NULL, *iqmp = NULL;
 
 #if USE_EVP
 	if (key->keydata.pkey == NULL)
@@ -1206,10 +1062,6 @@ opensslrsa_tofile(const dst_key_t *key, const char *directory) {
 #endif
 	memset(bufs, 0, sizeof(bufs));
 
-	RSA_get0_key(rsa, &n, &e, &d);
-	RSA_get0_factors(rsa, &p, &q);
-	RSA_get0_crt_params(rsa, &dmp1, &dmq1, &iqmp);
-
 	if (key->external) {
 		priv.nelements = 0;
 		result = dst__privstruct_writefile(key, &priv, directory);
@@ -1217,7 +1069,7 @@ opensslrsa_tofile(const dst_key_t *key, const char *directory) {
 	}
 
 	for (i = 0; i < 8; i++) {
-		bufs[i] = isc_mem_get(key->mctx, BN_num_bytes(n));
+		bufs[i] = isc_mem_get(key->mctx, BN_num_bytes(rsa->n));
 		if (bufs[i] == NULL) {
 			result = ISC_R_NOMEMORY;
 			goto fail;
@@ -1227,77 +1079,75 @@ opensslrsa_tofile(const dst_key_t *key, const char *directory) {
 	i = 0;
 
 	priv.elements[i].tag = TAG_RSA_MODULUS;
-	priv.elements[i].length = BN_num_bytes(n);
-	BN_bn2bin(n, bufs[i]);
+	priv.elements[i].length = BN_num_bytes(rsa->n);
+	BN_bn2bin(rsa->n, bufs[i]);
 	priv.elements[i].data = bufs[i];
 	i++;
 
 	priv.elements[i].tag = TAG_RSA_PUBLICEXPONENT;
-	priv.elements[i].length = BN_num_bytes(e);
-	BN_bn2bin(e, bufs[i]);
+	priv.elements[i].length = BN_num_bytes(rsa->e);
+	BN_bn2bin(rsa->e, bufs[i]);
 	priv.elements[i].data = bufs[i];
 	i++;
 
-	if (d != NULL) {
+	if (rsa->d != NULL) {
 		priv.elements[i].tag = TAG_RSA_PRIVATEEXPONENT;
-		priv.elements[i].length = BN_num_bytes(d);
-		BN_bn2bin(d, bufs[i]);
+		priv.elements[i].length = BN_num_bytes(rsa->d);
+		BN_bn2bin(rsa->d, bufs[i]);
 		priv.elements[i].data = bufs[i];
 		i++;
 	}
 
-	if (p != NULL) {
+	if (rsa->p != NULL) {
 		priv.elements[i].tag = TAG_RSA_PRIME1;
-		priv.elements[i].length = BN_num_bytes(p);
-		BN_bn2bin(p, bufs[i]);
+		priv.elements[i].length = BN_num_bytes(rsa->p);
+		BN_bn2bin(rsa->p, bufs[i]);
 		priv.elements[i].data = bufs[i];
 		i++;
 	}
 
-	if (q != NULL) {
+	if (rsa->q != NULL) {
 		priv.elements[i].tag = TAG_RSA_PRIME2;
-		priv.elements[i].length = BN_num_bytes(q);
-		BN_bn2bin(q, bufs[i]);
+		priv.elements[i].length = BN_num_bytes(rsa->q);
+		BN_bn2bin(rsa->q, bufs[i]);
 		priv.elements[i].data = bufs[i];
 		i++;
 	}
 
-	if (dmp1 != NULL) {
+	if (rsa->dmp1 != NULL) {
 		priv.elements[i].tag = TAG_RSA_EXPONENT1;
-		priv.elements[i].length = BN_num_bytes(dmp1);
-		BN_bn2bin(dmp1, bufs[i]);
+		priv.elements[i].length = BN_num_bytes(rsa->dmp1);
+		BN_bn2bin(rsa->dmp1, bufs[i]);
 		priv.elements[i].data = bufs[i];
 		i++;
 	}
 
-	if (dmq1 != NULL) {
+	if (rsa->dmq1 != NULL) {
 		priv.elements[i].tag = TAG_RSA_EXPONENT2;
-		priv.elements[i].length = BN_num_bytes(dmq1);
-		BN_bn2bin(dmq1, bufs[i]);
+		priv.elements[i].length = BN_num_bytes(rsa->dmq1);
+		BN_bn2bin(rsa->dmq1, bufs[i]);
 		priv.elements[i].data = bufs[i];
 		i++;
 	}
 
-	if (iqmp != NULL) {
+	if (rsa->iqmp != NULL) {
 		priv.elements[i].tag = TAG_RSA_COEFFICIENT;
-		priv.elements[i].length = BN_num_bytes(iqmp);
-		BN_bn2bin(iqmp, bufs[i]);
+		priv.elements[i].length = BN_num_bytes(rsa->iqmp);
+		BN_bn2bin(rsa->iqmp, bufs[i]);
 		priv.elements[i].data = bufs[i];
 		i++;
 	}
 
 	if (key->engine != NULL) {
 		priv.elements[i].tag = TAG_RSA_ENGINE;
-		priv.elements[i].length =
-			(unsigned short)strlen(key->engine) + 1;
+		priv.elements[i].length = strlen(key->engine) + 1;
 		priv.elements[i].data = (unsigned char *)key->engine;
 		i++;
 	}
 
 	if (key->label != NULL) {
 		priv.elements[i].tag = TAG_RSA_LABEL;
-		priv.elements[i].length =
-			(unsigned short)strlen(key->label) + 1;
+		priv.elements[i].length = strlen(key->label) + 1;
 		priv.elements[i].data = (unsigned char *)key->label;
 		i++;
 	}
@@ -1312,45 +1162,33 @@ opensslrsa_tofile(const dst_key_t *key, const char *directory) {
 	for (i = 0; i < 8; i++) {
 		if (bufs[i] == NULL)
 			break;
-		isc_mem_put(key->mctx, bufs[i], BN_num_bytes(n));
+		isc_mem_put(key->mctx, bufs[i], BN_num_bytes(rsa->n));
 	}
 	return (result);
 }
 
 static isc_result_t
-rsa_check(RSA *rsa, RSA *pub) {
-	const BIGNUM *n1 = NULL, *n2 = NULL;
-	const BIGNUM *e1 = NULL, *e2 = NULL;
-	BIGNUM *n = NULL, *e = NULL;
-
-	/*
-	 * Public parameters should be the same but if they are not set
-	 * copy them from the public key.
-	 */
-	RSA_get0_key(rsa, &n1, &e1, NULL);
+rsa_check(RSA *rsa, RSA *pub)
+{
+	/* Public parameters should be the same but if they are not set
+	 * copy them from the public key. */
 	if (pub != NULL) {
-		RSA_get0_key(pub, &n2, &e2, NULL);
-		if (n1 != NULL) {
-			if (BN_cmp(n1, n2) != 0)
+		if (rsa->n != NULL) {
+			if (BN_cmp(rsa->n, pub->n) != 0)
 				return (DST_R_INVALIDPRIVATEKEY);
 		} else {
-			n = BN_dup(n2);
+			rsa->n = pub->n;
+			pub->n = NULL;
 		}
-		if (e1 != NULL) {
-			if (BN_cmp(e1, e2) != 0)
+		if (rsa->e != NULL) {
+			if (BN_cmp(rsa->e, pub->e) != 0)
 				return (DST_R_INVALIDPRIVATEKEY);
 		} else {
-			e = BN_dup(e2);
-		}
-		if (RSA_set0_key(rsa, n, e, NULL) == 0) {
-			if (n != NULL)
-				BN_free(n);
-			if (e != NULL)
-				BN_free(e);
+			rsa->e = pub->e;
+			pub->e = NULL;
 		}
 	}
-	RSA_get0_key(rsa, &n1, &e1, NULL);
-	if (n1 == NULL || e1 == NULL)
+	if (rsa->n == NULL || rsa->e == NULL)
 		return (DST_R_INVALIDPRIVATEKEY);
 	return (ISC_R_SUCCESS);
 }
@@ -1362,17 +1200,13 @@ opensslrsa_parse(dst_key_t *key, isc_lex_t *lexer, dst_key_t *pub) {
 	int i;
 	RSA *rsa = NULL, *pubrsa = NULL;
 #ifdef USE_ENGINE
-	ENGINE *ep = NULL;
-	const BIGNUM *ex = NULL;
+	ENGINE *e = NULL;
 #endif
 	isc_mem_t *mctx = key->mctx;
 	const char *engine = NULL, *label = NULL;
 #if defined(USE_ENGINE) || USE_EVP
 	EVP_PKEY *pkey = NULL;
 #endif
-	BIGNUM *n = NULL, *e = NULL, *d = NULL;
-	BIGNUM *p = NULL, *q = NULL;
-	BIGNUM *dmp1 = NULL, *dmq1 = NULL, *iqmp = NULL;
 
 	/* read private key file */
 	ret = dst__privstruct_parse(key, DST_ALG_RSA, lexer, mctx, &priv);
@@ -1423,10 +1257,10 @@ opensslrsa_parse(dst_key_t *key, isc_lex_t *lexer, dst_key_t *pub) {
 #ifdef USE_ENGINE
 		if (engine == NULL)
 			DST_RET(DST_R_NOENGINE);
-		ep = dst__openssl_getengine(engine);
-		if (ep == NULL)
+		e = dst__openssl_getengine(engine);
+		if (e == NULL)
 			DST_RET(DST_R_NOENGINE);
-		pkey = ENGINE_load_private_key(ep, label, NULL, NULL);
+		pkey = ENGINE_load_private_key(e, label, NULL, NULL);
 		if (pkey == NULL)
 			DST_RET(dst__openssl_toresult2(
 					"ENGINE_load_private_key",
@@ -1442,8 +1276,7 @@ opensslrsa_parse(dst_key_t *key, isc_lex_t *lexer, dst_key_t *pub) {
 			DST_RET(dst__openssl_toresult(DST_R_OPENSSLFAILURE));
 		if (rsa_check(rsa, pubrsa) != ISC_R_SUCCESS)
 			DST_RET(DST_R_INVALIDPRIVATEKEY);
-		RSA_get0_key(rsa, NULL, &ex, NULL);
-		if (BN_num_bits(ex) > RSA_MAX_PUBEXP_BITS)
+		if (BN_num_bits(rsa->e) > RSA_MAX_PUBEXP_BITS)
 			DST_RET(ISC_R_RANGE);
 		if (pubrsa != NULL)
 			RSA_free(pubrsa);
@@ -1486,65 +1319,48 @@ opensslrsa_parse(dst_key_t *key, isc_lex_t *lexer, dst_key_t *pub) {
 			continue;
 		case TAG_RSA_LABEL:
 			continue;
-		case TAG_RSA_PIN:
-			continue;
 		default:
 			bn = BN_bin2bn(priv.elements[i].data,
 				       priv.elements[i].length, NULL);
 			if (bn == NULL)
 				DST_RET(ISC_R_NOMEMORY);
+		}
 
-			switch (priv.elements[i].tag) {
+		switch (priv.elements[i].tag) {
 			case TAG_RSA_MODULUS:
-				n = bn;
+				rsa->n = bn;
 				break;
 			case TAG_RSA_PUBLICEXPONENT:
-				e = bn;
+				rsa->e = bn;
 				break;
 			case TAG_RSA_PRIVATEEXPONENT:
-				d = bn;
+				rsa->d = bn;
 				break;
 			case TAG_RSA_PRIME1:
-				p = bn;
+				rsa->p = bn;
 				break;
 			case TAG_RSA_PRIME2:
-				q = bn;
+				rsa->q = bn;
 				break;
 			case TAG_RSA_EXPONENT1:
-				dmp1 = bn;
+				rsa->dmp1 = bn;
 				break;
 			case TAG_RSA_EXPONENT2:
-				dmq1 = bn;
+				rsa->dmq1 = bn;
 				break;
 			case TAG_RSA_COEFFICIENT:
-				iqmp = bn;
+				rsa->iqmp = bn;
 				break;
-			}
 		}
 	}
 	dst__privstruct_free(&priv, mctx);
 	memset(&priv, 0, sizeof(priv));
 
-	if (RSA_set0_key(rsa, n, e, d) == 0) {
-		if (n != NULL) BN_free(n);
-		if (e != NULL) BN_free(e);
-		if (d != NULL) BN_free(d);
-	}
-	if (RSA_set0_factors(rsa, p, q) == 0) {
-		if (p != NULL) BN_free(p);
-		if (q != NULL) BN_free(q);
-	}
-	if (RSA_set0_crt_params(rsa, dmp1, dmq1, iqmp) == 0) {
-		if (dmp1 != NULL) BN_free(dmp1);
-		if (dmq1 != NULL) BN_free(dmq1);
-		if (iqmp != NULL) BN_free(iqmp);
-	}
-
 	if (rsa_check(rsa, pubrsa) != ISC_R_SUCCESS)
 		DST_RET(DST_R_INVALIDPRIVATEKEY);
-	if (BN_num_bits(e) > RSA_MAX_PUBEXP_BITS)
+	if (BN_num_bits(rsa->e) > RSA_MAX_PUBEXP_BITS)
 		DST_RET(ISC_R_RANGE);
-	key->key_size = BN_num_bits(n);
+	key->key_size = BN_num_bits(rsa->n);
 	if (pubrsa != NULL)
 		RSA_free(pubrsa);
 #if USE_EVP
@@ -1578,12 +1394,12 @@ opensslrsa_fromlabel(dst_key_t *key, const char *engine, const char *label,
 	EVP_PKEY *pkey = NULL;
 	RSA *rsa = NULL, *pubrsa = NULL;
 	char *colon, *tmpengine = NULL;
-	const BIGNUM *ex = NULL;
 
 	UNUSED(pin);
 
 	if (engine == NULL) {
-		if (strchr(label, ':') == NULL)
+		colon = strchr(label, ':');
+		if (colon == NULL)
 			DST_RET(DST_R_NOENGINE);
 		tmpengine = isc_mem_strdup(key->mctx, label);
 		if (tmpengine == NULL)
@@ -1622,8 +1438,7 @@ opensslrsa_fromlabel(dst_key_t *key, const char *engine, const char *label,
 		DST_RET(dst__openssl_toresult(DST_R_OPENSSLFAILURE));
 	if (rsa_check(rsa, pubrsa) != ISC_R_SUCCESS)
 		DST_RET(DST_R_INVALIDPRIVATEKEY);
-	RSA_get0_key(rsa, NULL, &ex, NULL);
-	if (BN_num_bits(ex) > RSA_MAX_PUBEXP_BITS)
+	if (BN_num_bits(rsa->e) > RSA_MAX_PUBEXP_BITS)
 		DST_RET(ISC_R_RANGE);
 	if (pubrsa != NULL)
 		RSA_free(pubrsa);
@@ -1658,6 +1473,7 @@ opensslrsa_fromlabel(dst_key_t *key, const char *engine, const char *label,
 
 static dst_func_t opensslrsa_functions = {
 	opensslrsa_createctx,
+	NULL, /*%< createctx2 */
 	opensslrsa_destroyctx,
 	opensslrsa_adddata,
 	opensslrsa_sign,
